from oem_format_minimize.core.minimize import MinimizeProtocol


class NameMinimizeProtocol(MinimizeProtocol):
    __root__ = True
    __ignore__ = True

    __process__ = {
        'item': {
            'optional': True
        }
    }



from oem_framework.core.elapsed import Elapsed

from copy import deepcopy
import inspect


class MinimizeProtocol(object):
    __built = False
    __encode_map = {}
    __protocols = {}

    __key__ = None

    __ignore__ = False
    __process__ = None
    __root__ = False
    __version__ = None

    @classmethod
    def build(cls):
        if cls.__built:
            return

        cls.__decode_map = {}
        cls.__encode_map = {}

        cls.__protocols = {}

        for key in dir(cls):
            value = getattr(cls, key)

            if key.startswith('__') or key.startswith('_MinimizeProtocol__') or inspect.ismethod(value):
                continue

            if isinstance(value, MinimizeProperty):
                key = value.key
                value = value.value

            if type(value) is int:
                cls.__decode_map[value] = key
                cls.__encode_map[key] = value
            elif inspect.isclass(value):
                cls.__protocols[value.__key__] = value
            else:
                raise ValueError('Unknown property in protocol (key: %r)' % key)

        cls.__built = True

    @classmethod
    def get_protocol(cls, name):
        if name not in cls.__protocols:
            raise ValueError('Missing definition in %r for protocol %r' % (cls, name))

        return cls.__protocols[name]

    @classmethod
    def decode_key(cls, key):
        # Try retrieve decode mapping for `key`
        try:
            return cls.__decode_map[int(key)]
        except KeyError:
            raise ValueError('Missing definition in %r for key %r' % (cls, int(key)))

    @classmethod
    def encode_key(cls, key):
        if key not in cls.__encode_map:
            raise ValueError('Missing definition in %r for key %r' % (cls, key))

        return cls.__encode_map[key]

    @classmethod
    def get_value(cls, data, key, default=None):
        try:
            return data[str(key)]
        except KeyError:
            try:
                return data[key]
            except KeyError:
                return default

    @classmethod
    def to_child(cls, key=None, process=None):
        if process is not None:
            if type(process) is not dict:
                raise ValueError('Invalid value provided for "process" parameter, expected a dictionary')

            if cls.__process__:
                # Merge class `__process__` parameters with provided parameters
                result = {}
                result.update(deepcopy(cls.__process__))
                result.update(process)

                process = result

        class Child(cls):
            __key__ = key

            __process__ = process
            __root__ = False

        Child.__name__ = cls.__name__
        return Child


class MinimizeProperty(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value


class Minimize(object):
    @classmethod
    @Elapsed.track
    def decode(cls, data, protocol, process=None, children=True, ignore_keys=None):
        if protocol.__ignore__:
            return data

        # Pre-process `data`
        processed, result = cls._process(
            data, protocol,
            func=cls.decode,
            process=process
        )

        if processed:
            return result

        if ignore_keys is None:
            ignore_keys = []

        # Ensure protocol is built
        protocol.build()

        # Process each item in `data`
        for key in list(data.keys()):
            # Ignore "minimized" item identifier
            if key == '~':
                data.pop(key)
                continue

            # Ignore keys defined in `ignore_keys`
            if type(key) is str and key in ignore_keys:
                continue

            # Decode `key` with `protocol`
            value = data.pop(key)
            key = protocol.decode_key(key)

            if children:
                # Process child protocols
                value = cls._decode_children(key, value, protocol)

            # Store item
            data[key] = value

        return data

    @classmethod
    def _decode_children(cls, key, value, protocol):
        if type(value) is dict:
            # Decode `value` with `protocol`
            return cls.decode(value, protocol.get_protocol(key))

        if type(value) is list:
            # Decode items in `value` with `protocol`
            return [
                (
                    cls.decode(value, protocol.get_protocol(key))
                    if type(value) is dict else value
                )
                for value in value
            ]

        return value

    @classmethod
    @Elapsed.track
    def encode(cls, data, protocol, process=None):
        if protocol.__ignore__:
            return data

        # Pre-process `data`
        processed, result = cls._process(
            data, protocol,
            func=cls.encode,
            process=process
        )

        if processed:
            return result

        # Ensure protocol is built
        protocol.build()

        # Process each item in `data`
        result = {}

        for key, value in data.items():
            # Process child protocols
            if type(value) is dict:
                # Encode `value` with `protocol`
                value = cls.encode(value, protocol.get_protocol(key))
            elif type(value) is list:
                # Encode items in `value` with `protocol`
                value = [
                    (
                        cls.encode(value, protocol.get_protocol(key))
                        if type(value) is dict else value
                    )
                    for value in value
                ]

            # Encode `key` with `protocol`
            key = protocol.encode_key(key)

            # Store item
            result[key] = value

        if protocol.__root__:
            if protocol.__version__ is None:
                raise ValueError('Missing "__version__" parameter on %r' % protocol)

            # Set "minimize" flag (and version)
            result['~'] = protocol.__version__

        return result

    @classmethod
    @Elapsed.track
    def _process(cls, data, protocol, func, process=None):
        # Retrieve parameters
        mode, optional, supported = cls._process_parameters(protocol, process)

        # Pre-process `data`
        if mode == 'children':
            # Process dictionary of children
            if type(data) is dict and (supported is None or type(data) in supported):
                cls._transform_dictionary(data, lambda v: func(v, protocol, process='item'))
                return True, data

            # Process list of children
            if type(data) is list and (supported is None or type(data) in supported):
                return True, [
                    func(value, protocol, process='item')
                    for value in data
                ]

            if not optional:
                raise ValueError('Dictionary or List required for %r protocol' % protocol)
        elif mode == 'item':
            # Process dictionary of item children
            if type(data) is dict and supported and type(data) in supported:
                cls._transform_dictionary(data, lambda v: func(v, protocol, process='item'))
                return True, data

            # Process list of item children
            if type(data) is list and supported and type(data) in supported:
                return True, [
                    func(value, protocol, process='item')
                    for value in data
                ]

        # No processing required
        return False, None

    @classmethod
    @Elapsed.track
    def _process_parameters(cls, protocol, process):
        mode = None

        if process is None:
            process = protocol.__process__

            if type(process) is dict and process and 'children' in process:
                mode = 'children'

                if type(process) is not bool:
                    process = process['children']
            elif type(process) is str:
                mode = process
            elif process:
                raise ValueError

        if type(process) is str:
            if type(protocol.__process__) is dict and process in protocol.__process__:
                mode = process
                process = protocol.__process__[process]
            else:
                mode = process

        # Retrieve parameters
        if type(process) is dict:
            return (
                mode or process.get('mode'),
                process.get('optional', False),
                process.get('supported')
            )

        return (
            mode or process,
            False,
            None
        )

    @classmethod
    @Elapsed.track
    def _transform_dictionary(cls, data, func):
        for key in data.iterkeys():
            data[key] = func(data[key])



from oem_format_minimize.core.minimize import MinimizeProtocol
from oem_format_minimize.protocol.metadata import MetadataMinimizeProtocol


class IndexMinimizeProtocol(MinimizeProtocol):
    __root__    = True
    __version__ = 0x01

    items       = 0x01

    MetadataMinimizeProtocol = MetadataMinimizeProtocol.to_child(
        key='items',
        process={
            'children': True
        }
    )



from oem_format_minimize.protocol.index import IndexMinimizeProtocol
from oem_format_minimize.protocol.metadata import MetadataMinimizeProtocol

from oem_format_minimize.protocol.show import ShowMinimizeProtocol
from oem_format_minimize.protocol.season import SeasonMinimizeProtocol
from oem_format_minimize.protocol.episode import EpisodeMinimizeProtocol

from oem_format_minimize.protocol.movie import MovieMinimizeProtocol



from oem_format_minimize.core.minimize import MinimizeProtocol


class MetadataMinimizeProtocol(MinimizeProtocol):
    __root__ = True

    hashes      = 0x01
    media       = 0x02

    created_at  = 0x11
    updated_at  = 0x12

    class HashesProtocol(MinimizeProtocol):
        __key__ = 'hashes'
        __ignore__ = True



from oem_format_minimize.core.minimize import Minimize
from oem_format_minimize.protocol import *
from oem_framework.models import *
from oem_framework.format import Format

import inspect

# Setup model protocols
Index.set_protocol('minimize', IndexMinimizeProtocol)
Metadata.set_protocol('minimize', MetadataMinimizeProtocol)

Movie.set_protocol('minimize', MovieMinimizeProtocol)
Show.set_protocol('minimize', ShowMinimizeProtocol)
Season.set_protocol('minimize', SeasonMinimizeProtocol)
Episode.set_protocol('minimize', EpisodeMinimizeProtocol)


class MinimalFormat(Format):
    def encode(self, model, data, **kwargs):
        # Retrieve item minimize protocol
        protocol = model.__protocols__.get('minimize') if model.__protocols__ else None

        if protocol is None:
            raise ValueError('Model %r has no "minimize" protocol defined' % model)

        if inspect.isfunction(protocol):
            protocol = protocol()

        return Minimize.encode(data, protocol)

    def decode(self, model, encoded, children=True, ignore_keys=None, **kwargs):
        if model.__wrapper__:
            media = kwargs.get('media')

            if media is None:
                raise ValueError('Missing required parameters: "media"')

            model = model.resolve(media)

        # Retrieve item minimize protocol
        protocol = model.__protocols__.get('minimize') if model.__protocols__ else None

        if protocol is None:
            raise ValueError('Model %r has no "minimize" protocol defined' % model)

        if inspect.isfunction(protocol):
            protocol = protocol()

        return Minimize.decode(
            encoded, protocol,
            children=children,
            ignore_keys=ignore_keys
        )

    def from_dict(self, collection, model, encoded, children=True, **kwargs):
        # Decode dictionary with minimized data protocol
        data = self.decode(model, encoded, children=children, **kwargs)

        # Parse `model` from `data`
        return Format.from_dict(self, collection, model, data, **kwargs)

    def to_dict(self, item, **kwargs):
        # Convert `item` to `data` dictionary
        data = Format.to_dict(self, item, **kwargs)

        # Encode dictionary with minimized data protocol
        return self.encode(item.__class__, data)



from oem_format_minimize.core.minimize import MinimizeProtocol, MinimizeProperty
from oem_format_minimize.protocol.name import NameMinimizeProtocol


class PartMinimizeProtocol(MinimizeProtocol):
    __root__ = True
    __process__ = {
        'item': {
            'optional': True,
            'supported': [
                list
            ]
        }
    }

    number          = 0x01

    identifiers     = 0x11
    names           = 0x12

    supplemental    = 0x21
    parameters      = 0x22

    NameMinimizeProtocol = NameMinimizeProtocol.to_child(
        key='names',
        process={
            'children': True
        }
    )

    class IdentifiersMinimizeProtocol(MinimizeProtocol):
        __key__ = 'identifiers'

        anidb           = 0x01
        imdb            = 0x02
        tvdb            = 0x03
        tmdb_movie      = MinimizeProperty('tmdb:movie', 0x04)
        tmdb_show       = MinimizeProperty('tmdb:show',  0x05)

    class SupplementalMinimizeProtocol(MinimizeProtocol):
        __key__ = 'supplemental'

        studio          = 0x01

    class ParametersMinimizeProtocol(MinimizeProtocol):
        __key__ = 'parameters'

        default_season  = 0x01
        episode_offset  = 0x02



from oem_format_minimize.core.minimize import MinimizeProtocol, MinimizeProperty
from oem_format_minimize.protocol.episode import EpisodeMinimizeProtocol
from oem_format_minimize.protocol.name import NameMinimizeProtocol


class SeasonMinimizeProtocol(MinimizeProtocol):
    __root__ = True

    identifiers     = 0x01
    names           = 0x02

    supplemental    = 0x11
    parameters      = 0x12

    mappings        = 0x21

    episodes        = 0x31

    EpisodeMinimizeProtocol = EpisodeMinimizeProtocol.to_child(
        key='episodes',
        process={
            'children': True
        }
    )

    NameMinimizeProtocol = NameMinimizeProtocol.to_child(
        key='names',
        process={
            'children': True
        }
    )

    class IdentifiersMinimizeProtocol(MinimizeProtocol):
        __key__ = 'identifiers'

        anidb           = 0x01
        imdb            = 0x02
        tvdb            = 0x03
        tmdb_movie      = MinimizeProperty('tmdb:movie', 0x04)
        tmdb_show       = MinimizeProperty('tmdb:show',  0x05)

    class SupplementalMinimizeProtocol(MinimizeProtocol):
        __key__ = 'supplemental'

        studio          = 0x01

    class ParametersMinimizeProtocol(MinimizeProtocol):
        __key__ = 'parameters'

        default_season  = 0x01
        episode_offset  = 0x02

    class SeasonMappingMinimizeProtocol(MinimizeProtocol):
        __key__ = 'mappings'

        identifiers = 0x01
        names       = 0x02

        season      = 0x11

        start       = 0x21
        end         = 0x22
        offset      = 0x23

        class IdentifiersMinimizeProtocol(MinimizeProtocol):
            __key__ = 'identifiers'

            anidb           = 0x01
            imdb            = 0x02
            tvdb            = 0x03
            tmdb_movie      = MinimizeProperty('tmdb:movie', 0x04)
            tmdb_show       = MinimizeProperty('tmdb:show',  0x05)

        NameMinimizeProtocol = NameMinimizeProtocol.to_child(
            key='names',
            process={
                'children': True
            }
        )
